"""
Lambada command-line interface
"""

import argparse
import os
from typing import Any, Callable

from simiotics.cli import read_string_from_file
from simiotics.client import client_from_env, Simiotics

from . import handlers

LambdaBasicExecutionRolePolicy = """
{
    "Version": "2012-10-17",
    "Statement": [
        {
            "Effect": "Allow",
            "Action": [
                "logs:CreateLogGroup",
                "logs:CreateLogStream",
                "logs:PutLogEvents"
            ],
            "Resource": "*"
        }
    ]
}
""".strip()

def generate_cli() -> argparse.ArgumentParser:
    """
    Generates the lambada CLI

    Args: None

    Returns: argparse.ArgumentParser object representing the lambada CLI
    """
    description = """
lambada: Manage AWS Lambdas using Simiotics Function Registries
\n
Note: This tool uses the SIMIOTICS_FUNCTION_REGISTRY environment variable to determine the Function
Registry to communicate with. If the variable is not set, it will error out. You can use the public
Simiotics Function Registry by running:
```
$ export SIMIOTICS_FUNCTION_REGISTRY=registry-alpha.simiotics.com:7011
```
    """
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    subparsers = parser.add_subparsers(title='Commands')

    register = subparsers.add_parser(
        'register',
        help='Register a lambada function against a Simiotics Function Registry and print its key',
    )
    register.add_argument(
        '--runtime',
        choices=['python3.6', 'python3.7'],
        default='python3.7',
        help='Python runtime to use for the given function (default: python3.7)',
    )
    register.add_argument(
        '-k',
        '--key',
        type=str,
        required=True,
        help='Key under which function should be registered against function registry',
    )
    register.add_argument(
        '-c',
        '--code',
        type=read_string_from_file,
        required=True,
        help='Path to file containing the function code',
    )
    register.add_argument(
        '--handler',
        type=str,
        required=True,
        help='Python specification of handler',
    )
    register.add_argument(
        '--requirements',
        type=read_string_from_file,
        default='',
        help='Path to file specifying requirements for the function (optional)',
    )
    register.add_argument(
        '--iam-policy',
        type=str,
        default=LambdaBasicExecutionRolePolicy,
        help=(
            'IAM policy that should be granted to an AWS Lambda running the given function '
            '(default: allows Lambda to log usage in CloudWatch)'
        ),
    )
    register.add_argument(
        '--timeout',
        type=int,
        default=3,
        help='Timeout for execution of Lambda in seconds (default: 3)',
    )
    register.add_argument(
        '--overwrite',
        action='store_true',
        help='If a function has already been registered under the given key, overwrite it',
    )
    register.set_defaults(func=handlers.register)

    create_role = subparsers.add_parser(
        'create_role',
        help=(
            'Create IAM role to be assumed by AWS Lambda deployment of function from Simiotics '
            'Function Registry and print its name'
        )
    )
    create_role.add_argument(
        '-k',
        '--key', type=str,
        required=True,
        help='Key of function in Simiotics Function Registry for which to create IAM role',
    )
    create_role.add_argument(
        '-n',
        '--name',
        type=str,
        help='Name for IAM role',
    )
    create_role.set_defaults(func=handlers.create_role)

    deploy = subparsers.add_parser(
        'deploy',
        help='Deploy a Simiotics Function as an AWS Lambda and print its AWS Lambda ARN',
    )
    deploy.add_argument(
        '-n',
        '--name',
        type=str,
        default=None,
        help='Name of the AWS Lambda to deploy',
    )
    deploy.add_argument(
        '-k',
        '--key',
        type=str,
        required=True,
        help='Key for function in Simiotics Function Registry',
    )
    deploy.add_argument(
        '--keep-staging-dir',
        action='store_true',
        help='If set, keeps the deployment package temporary directory. Path is printed to stderr.'
    )
    deploy.set_defaults(func=handlers.deploy)

    down = subparsers.add_parser(
        'down',
        help='Take down an AWS Lambda deployed via lambada',
    )
    down.add_argument(
        '-k',
        '--key',
        type=str,
        required=True,
        help='Key for function in Simiotics Function Registry',
    )
    down.add_argument(
        '--teardown',
        action='store_true',
        help='All AWS resources associated with the Simiotics function should be removed'
    )
    down.set_defaults(func=handlers.down)

    list_functions = subparsers.add_parser(
        'list',
        help='List functions in the Simiotics Function Registry managed by lambada',
    )
    list_functions.add_argument(
        '--num-items',
        type=int,
        default=100,
        help='Number of registry functions that should be processed at a time (default: 100).'
    )
    list_functions.set_defaults(func=handlers.list_functions)

    return parser

def main() -> None:
    """
    Runs the lambada tool

    Args: None

    Returns: None
    """

    if os.environ.get('SIMIOTICS_FUNCTION_REGISTRY') is None:
        raise ValueError('SIMIOTICS_FUNCTION_REGISTRY environment variable undefined')

    parser = generate_cli()
    args = parser.parse_args()
    args.func(args)

if __name__ == '__main__':
    main()



"""
Handlers for lambada commands
"""

import argparse
import os
import shutil
import subprocess
import sys
import tempfile

import boto3
from simiotics.client import client_from_env

AWSLambdaTrustedEntity = """
{
    "Version": "2012-10-17",
    "Statement": [
        {
        "Sid": "",
        "Effect": "Allow",
        "Principal": {
            "Service": "lambda.amazonaws.com"
        },
        "Action": "sts:AssumeRole"
        }
    ]
}
""".strip()

SimioticsPath = '/simiotics/'

LambadaManagerKey = 'manager'
LambadaManager = 'lambada'

def register(args: argparse.Namespace) -> None:
    """
    Handler for `lambada register`, which registers a new lambada function against a simiotics
    registry

    Args:
    args
        `argparse.Namespace` object containing parameters to the `register` command

    Returns: None, prints key of registered function
    """
    simiotics = client_from_env()

    tags = {
        'runtime': args.runtime,
        'handler': args.handler,
        'requirements': args.requirements,
        'iam_policy': args.iam_policy,
        'timeout': str(args.timeout),
        LambadaManagerKey: LambadaManager,
    }

    simiotics.register_function(args.key, args.code, tags, args.overwrite)

    print(args.key)

def create_role(args: argparse.Namespace) -> None:
    """
    Handler for `lambada create_role`, which creates an AWS IAM role that an AWS Lambda implementing
    the specified Simiotics Function Registry function can use in its execution

    Args:
    args
        `argparse.Namespace` object containing parameters to the `create_role` command

    Returns: None, prints IAM role name
    """
    simiotics = client_from_env()
    registered_function = simiotics.get_registered_function(args.key)
    if registered_function.tags.get(LambadaManagerKey) != LambadaManager:
        raise ValueError('Simiotics function with key={} not managed by lambada'.format(args.key))

    iam_client = boto3.client('iam')

    response = iam_client.create_role(
        Path=SimioticsPath,
        RoleName=args.name,
        AssumeRolePolicyDocument=AWSLambdaTrustedEntity,
        Description='AWS Lambda execution role for Simiotics function: {}'.format(args.key),
        Tags=[
            {'Key': 'Creator', 'Value': 'simiotics'},
        ]
    )

    role_name = response['Role']['RoleName']
    role_arn = response['Role']['Arn']

    iam_policy = registered_function.tags['iam_policy']
    policy_name = 'lambada_policy_{}'.format(role_name)
    iam_client.put_role_policy(
        RoleName=role_name,
        PolicyName=policy_name,
        PolicyDocument=iam_policy,
    )

    tags = registered_function.tags
    tags['iam_role_name'] = role_name
    tags['iam_role_arn'] = role_arn
    tags['iam_role_policy'] = policy_name
    simiotics.register_function(
        key=registered_function.key,
        code=registered_function.code,
        tags=tags,
        overwrite=True,
    )

    print(role_name)

def deploy(args: argparse.Namespace) -> None:
    """
    Handler for `lambada deploy`, which creates an AWS Lambda from a Simiotics function

    Args:
    args
        `argparse.Namespace` object containing parameters to the `deploy` command

    Returns: None, prints AWS Lambda ARN
    """
    simiotics = client_from_env()
    registered_function = simiotics.get_registered_function(args.key)
    if registered_function.tags.get(LambadaManagerKey) != LambadaManager:
        raise ValueError('Simiotics function with key={} not managed by lambada'.format(args.key))

    staging_dir = tempfile.mkdtemp()
    try:
        deployment_package_dir = os.path.join(staging_dir, 'deployment_package')
        os.mkdir(deployment_package_dir)
        requirements_txt = os.path.join(staging_dir, 'requirements.txt')
        code_py = os.path.join(deployment_package_dir, 'code.py')

        with open(requirements_txt, 'w') as ofp:
            ofp.write(registered_function.tags['requirements'])

        subprocess.run(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "-r",
                requirements_txt,
                "--target",
                deployment_package_dir
            ],
            check=True,
        )
        if os.path.exists(code_py):
            raise ValueError('File already exists at path: {}'.format(code_py))

        with open(code_py, 'w') as ofp:
            ofp.write(registered_function.code)

        zipfilepath = os.path.join(staging_dir, 'function.zip')
        shutil.make_archive(os.path.splitext(zipfilepath)[0], 'zip', deployment_package_dir)
        with open(zipfilepath, 'rb') as ifp:
            deployment_package = ifp.read()

        lambda_client = boto3.client('lambda')
        handler_path = 'code.{}'.format(registered_function.tags['handler'])
        lambda_resource = lambda_client.create_function(
            FunctionName=args.name,
            Runtime=registered_function.tags['runtime'],
            Role=registered_function.tags['iam_role_arn'],
            Handler=handler_path,
            Code={'ZipFile': deployment_package},
            Description='Simiotics lambada deployment of: {}'.format(args.key),
            Timeout=int(registered_function.tags['timeout']),
            Tags={
                'Creator': 'simiotics',
            },
        )

        lambda_arn = lambda_resource['FunctionArn']

        registered_function.tags['lambda_arn'] = lambda_arn
        simiotics.register_function(
            key=registered_function.key,
            code=registered_function.code,
            tags=registered_function.tags,
            overwrite=True,
        )
        print(lambda_arn)
    finally:
        if not args.keep_staging_dir:
            shutil.rmtree(staging_dir)
        else:
            print(staging_dir, file=sys.stderr)

def down(args: argparse.Namespace) -> None:
    """
    Handler for `lambada down`, which takes down an AWS Lambda from a Simiotics function

    Args:
    args
        `argparse.Namespace` object containing parameters to the `down` command

    Returns: None, prints Simiotics Function Registry key
    """
    simiotics = client_from_env()
    registered_function = simiotics.get_registered_function(args.key)
    if registered_function.tags.get(LambadaManagerKey) != LambadaManager:
        raise ValueError('Simiotics function with key={} not managed by lambada'.format(args.key))

    try:
        lambda_arn = registered_function.tags.get('lambda_arn')
        if lambda_arn is not None:
            lambda_client = boto3.client('lambda')
            lambda_client.delete_function(FunctionName=lambda_arn)
            registered_function.tags.pop('lambda_arn')

        if args.teardown:
            iam_client = boto3.client('iam')
            role_name = registered_function.tags.get('iam_role_name')
            if role_name is not None:
                policy_name = registered_function.tags.get('iam_role_policy')
                if policy_name is not None:
                    iam_client.delete_role_policy(RoleName=role_name, PolicyName=policy_name)
                    registered_function.tags.pop('iam_role_policy')

                iam_client.delete_role(RoleName=role_name)
                registered_function.tags.pop('iam_role_name')
                registered_function.tags.pop('iam_role_arn')
    finally:
        simiotics.register_function(
            key=registered_function.key,
            code=registered_function.code,
            tags=registered_function.tags,
            overwrite=True,
        )

    print(args.key)

def list_functions(args: argparse.Namespace) -> None:
    """
    Handler for `lambada list`, which lists all the functions in a Simiotics Function Registry which
    are managed by lambada. The function registry for which functions are listed is the one
    specified by the SIMIOTICS_FUNCTION_REGISTRY environment variable.

    Args:
    args
        `argparse.Namespace` object containing parameters to the `list` command

    Returns: None, prints lambada-managed functions in the Simiotics Function Registry line by line
    """
    simiotics = client_from_env()

    proceed = True
    current_offset = 0
    while proceed:
        registered_functions = simiotics.list_registered_functions(current_offset, args.num_items)
        for registered_function in registered_functions:
            if registered_function.tags.get(LambadaManagerKey) == LambadaManager:
                print(registered_function.key)

        if len(registered_functions) < args.num_items:
            proceed = False

        current_offset += len(registered_functions)



from setuptools import find_packages, setup

VERSION = '0.1.1'

long_description = ''
with open('README.md') as ifp:
    long_description = ifp.read()

setup(
    name='simiotics-lambada',
    version=VERSION,
    packages=find_packages(),
    install_requires=[
        'boto3',
        'simiotics',
    ],
    description='Manage AWS Lambda functions using a Simiotics Function Registry',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Neeraj Kashyap',
    author_email='neeraj@simiotics.com',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python',
    ],
    url='https://github.com/simiotics/lambada',
    entry_points={
        'console_scripts': [
            'lambada = lambada.cli:main'
        ]
    }
)
